/* $Id: hawki_bkg.c,v 1.15 2009/10/23 15:13:06 cgarcia Exp $
 *
 * This file is part of the HAWKI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: cgarcia $
 * $Date: 2009/10/23 15:13:06 $
 * $Revision: 1.15 $
 * $Name: hawki-1_8_0 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <string.h>
#include <cpl.h>

#include "hawki_distortion.h"
#include "hawki_bkg.h"
#include "hawki_pfits.h"
#include "hawki_load.h"
#include "hawki_utils.h"

/*----------------------------------------------------------------------------*/
/**
 * @defgroup hawki_image_stats   Compute several statistics on images 
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Fill a propertylist with the association between objects and bkg 
  @param    objframes   The objframes associated to the background
  @param    proplist    The list of keywords to store the association
                        information to  
  @return   0 if everything is ok, -1 otherwise

  This function takes as an input a set of object frames. 
  For EACH input object frames it creates in the given propertylist a
  keyword "pointing" to the ARCFILE of that frame. 
  This propertylist can later be stored in the background image to establish 
  the association between this background and the object frames. 
 */
/*----------------------------------------------------------------------------*/
int hawki_bkg_fill_assoc(cpl_frameset * objframes, cpl_propertylist * proplist)
{
    int   iframe;
    int   nframes;
    
    /* Loop on the frames */
    nframes = cpl_frameset_get_size(objframes);
    for(iframe = 0 ; iframe < nframes ; ++iframe)
    {
        cpl_frame         * this_frame;
        cpl_propertylist  * this_proplist;
        const char        * arcfile;
        char                keystr[256];

        /* Getting the proplist for this frame */
        this_frame     = cpl_frameset_get_frame(objframes, iframe);
        this_proplist  = cpl_propertylist_load
            (cpl_frame_get_filename(this_frame),0);
        arcfile        = hawki_pfits_get_arcfile(proplist);
        snprintf(keystr, 256, "ESO QC BKG ASSOC RAW%d",iframe+1);
        cpl_propertylist_append_string(proplist, keystr, arcfile);
        cpl_propertylist_delete(this_proplist);
    }
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Computes the median of the object frames 
  @param    objframes   The objframes to get the background from
  @param    bkg         The background image
  @param    proplist    The list of keywords to store the association
                        information to  
  @return   0 if everything is ok, -1 otherwise

  This function takes as an input a set of object frames and computes
  the background using a median of these frames.
 */
/*----------------------------------------------------------------------------*/
int hawki_bkg_from_objects_median
(const cpl_frameset  *  objframes, cpl_imagelist * bkg)
{
    int             idet;

    /* Error state variables */
    cpl_errorstate  prestate = cpl_errorstate_get();
    
    /* Loop on detectors */
    for(idet = 0; idet < HAWKI_NB_DETECTORS ; ++idet)
    {
        cpl_imagelist * img_serie;
        cpl_image     * this_bkg_image;
        
        /* Loading the object frame */
        img_serie = hawki_load_detector(objframes, idet + 1, CPL_TYPE_FLOAT);
        if(img_serie== NULL)
        {
            cpl_msg_error(__func__, "Error reading object image") ;
            return -1;
        }
        
        /* Averaging */
        /* TODO: Comprobar que cpl_imagelist_collapse_median_create
         * admite las mascaras, para que cuando implemente lo de la mascara 
         * de objetos siga funcionando */
        if ((this_bkg_image = cpl_imagelist_collapse_median_create(img_serie))
                == NULL)
        {
            cpl_msg_error(__func__, "Cannot compute the median of obj images");
            cpl_imagelist_delete(img_serie);
            return -1;
        }

        /* Fill the corresponding chip */
        cpl_imagelist_set(bkg, this_bkg_image, idet);
        
        /* Freeing */
        cpl_imagelist_delete(img_serie);
    }

    /* Check error status and exit */
    if(!cpl_errorstate_is_equal(prestate))
        return -1;
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Computes the median of the sky frames 
  @param    skyframes   The skyframes to get the background from
  @param    objframes   The objframes to which the association is made
  @param    bkg         The background image
  @param    proplist    The list of keywords to store the association
                        information to  
  @return   0 if everything is ok, -1 otherwise

  This function takes as an input a set of sky frames and computes
  the background using a median of these frames.
 */
/*----------------------------------------------------------------------------*/
int hawki_bkg_from_sky_median
(const cpl_frameset  *  skyframes, 
 cpl_imagelist       *  bkg)
{
    int             idet;

    /* Error state variables */
    cpl_errorstate  prestate = cpl_errorstate_get();
    
    /* Loop on detectors */
    for(idet = 0; idet < HAWKI_NB_DETECTORS ; ++idet)
    {
        cpl_imagelist * img_serie;
        cpl_image     * this_bkg_image;
        
        /* Loading the object frame */
        img_serie = hawki_load_detector(skyframes, idet + 1, CPL_TYPE_FLOAT);
        if(img_serie== NULL)
        {
            cpl_msg_error(__func__, "Error reading object image") ;
            return -1;
        }
        
        /* Averaging */
        if ((this_bkg_image = cpl_imagelist_collapse_median_create(img_serie))
                == NULL) 
        {
            cpl_msg_error(__func__, "Cannot compute the median of obj images");
            cpl_imagelist_delete(img_serie);
            return -1;
        }
        /* Fill the corresponding chip */
        cpl_imagelist_set(bkg, this_bkg_image, idet);
        
        /* Freeing */
        cpl_imagelist_delete(img_serie);
    }

    /* Check error status and exit */
    if(!cpl_errorstate_is_equal(prestate))
        return -1;
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief  Computes the running mean for a set of object images using their masks  
  @param    objimages   The serie of object images
  @param    medians     A vector with the medians of each image
  @param    bkg         The background image (OUT)
  @param    proplist    The list of keywords to store the association
                        information to  
  @return   0 if everything is ok, -1 otherwise

  This function takes as an input a set of sky frames and computes
  the background using a median of these frames. The bad pixels from the 
  bpm masks are taking into account. It also sets the association
  between the object images and the final background produced
  (calling hawki_bkg_fill_assoc).
 */
/*----------------------------------------------------------------------------*/
int hawki_bkg_from_running_mean_detector
(cpl_imagelist       * objimages,
 const cpl_vector    * medians,
 int                   i_target,
 int                   half_width,
 int                   rejlow,
 int                   rejhigh,
 cpl_image          *  bkg)
{
    float        *  bkg_p;
    float        *  curima_p;
    cpl_binary   *  curmask_p;
    float        ** objimages_p;
    cpl_binary   ** maskimages_p;
    const double *  medians_p;
    double          out;
    int             from_ima;
    int             to_ima;
    int             nima;
    int             nuse;
    int             nx;
    int             ny;
    
    int             pos_x;
    int             pos_y;
    int             pos;
    int             i_ima;
    int             i_win;
    int             n_win;
    
    cpl_vector * localwin;

    /* Get the pointer to the background image */
    bkg_p = cpl_image_get_data_float(bkg);
    
    /* Compute border indices */
    nima = cpl_imagelist_get_size(objimages);
    from_ima = i_target - half_width;
    to_ima   = i_target + half_width;
    if (from_ima<0) from_ima = 0 ;
    if (to_ima>(nima-1)) to_ima=nima-1 ;

    /* Number of valid planes to consider after edge effects */
    nuse = to_ima - from_ima ;

    /* Get the size of images */
    nx = cpl_image_get_size_x(bkg);
    ny = cpl_image_get_size_y(bkg);
    
    /* Get the pointers for better performance */
    medians_p = cpl_vector_get_data_const(medians);

    /* Allocate local window with the maximum number of possible used values*/
    localwin = cpl_vector_new(nuse) ;
    
    /* Get fast accesors to input image pointers */
    objimages_p  = cpl_malloc(nima * sizeof(float *));
    maskimages_p = cpl_malloc(nima * sizeof(cpl_binary *));
    for(i_ima=from_ima ; i_ima<=to_ima ; i_ima++)
    {
        objimages_p[i_ima]  = cpl_image_get_data_float
            (cpl_imagelist_get(objimages, i_ima));
        maskimages_p[i_ima] = cpl_mask_get_data
            (cpl_image_get_bpm(cpl_imagelist_get(objimages, i_ima)));
    }
    
    /* Initialize the bad pixel mask */
    cpl_image_accept_all(bkg);

    /* Loop over all pixels */
    for (pos_x=0 ; pos_x<nx ; pos_x++)
    {
        for (pos_y=0 ; pos_y<ny ; pos_y++)
        {
            /* Get the position in the array */
            pos = pos_x + pos_y * nx;
            
            /* Resize again to maximum number of elements */
            cpl_vector_set_size(localwin, nuse);
        
            /* Fill up local window */
            i_win=0;
            for (i_ima=from_ima ; i_ima<=to_ima ; i_ima++) 
            {
                if (i_ima != i_target) 
                {
                    curima_p = objimages_p[i_ima];
                    curmask_p = maskimages_p[i_ima];
                    if(!curmask_p[pos])
                    {
                        cpl_vector_set(localwin, i_win,
                                       (double)curima_p[pos]-medians_p[i_ima]);
                        i_win++;
                    }
                }
            }
            n_win = i_win;
            if(n_win - rejlow - rejhigh < 1)
            {
                /* Add it to the bad pixel mask */
                cpl_msg_debug(__func__,"Pixel %d %d added to the sky bpm",
                        pos_x, pos_y);
                if(cpl_image_reject(bkg, pos_x + 1, pos_y + 1) != CPL_ERROR_NONE)
                {
                    cpl_msg_error(__func__,"Cannot add pixel to sky bpm");
                    cpl_vector_delete(localwin) ;
                    cpl_free(objimages_p);
                    cpl_free(maskimages_p);
                    return(-1);
                }
            }
            else
            {
                /* Resize to current number of elements */
                cpl_vector_set_size(localwin, n_win);
                /* Sort window */
                cpl_vector_sort(localwin, 1);
                /* Reject min and max, accumulate other pixels */
                out = 0.0 ;
                for (i_win=rejlow ; i_win<n_win - rejhigh; i_win++) 
                {
                    out += cpl_vector_get(localwin, i_win);
                }
                /* Take the mean */
                out /= (double)(n_win - rejlow - rejhigh);
        
                /* Assign value */
                bkg_p[pos] = (out+medians_p[i_target]);
            }
        }
    }

    /* Free */
    cpl_vector_delete(localwin) ;
    cpl_free(objimages_p);
    cpl_free(maskimages_p);

    return 0;
}

/**@}*/

