/* $Id: hawki_cal_illum.c,v 1.8 2010/06/04 09:45:43 cgarcia Exp $
 *
 * This file is part of the HAWKI Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: cgarcia $
 * $Date: 2010/06/04 09:45:43 $
 * $Revision: 1.8 $
 * $Name: hawki-1_8_0 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <math.h>
#include <cpl.h>

#include "irplib_utils.h"
#include "irplib_strehl.h"

#include "hawki_utils.h"
#include "hawki_calib.h"
#include "hawki_pfits.h"
#include "hawki_dfs.h"
#include "hawki_load.h"
#include "hawki_save.h"

/*-----------------------------------------------------------------------------
                            Functions prototypes
 -----------------------------------------------------------------------------*/

static int hawki_cal_illum_create(cpl_plugin *) ;
static int hawki_cal_illum_exec(cpl_plugin *) ;
static int hawki_cal_illum_destroy(cpl_plugin *) ;
static int hawki_cal_illum(cpl_parameterlist *, cpl_frameset *) ;
static int hawki_cal_illum_save(const cpl_image **, const cpl_table **, 
        cpl_parameterlist *, cpl_frameset *) ;
static cpl_image * hawki_cal_illum_compute(cpl_frameset *, int, 
        const char *, const char *, cpl_table **) ;
static cpl_vector * hawki_cal_illum_phot(cpl_imagelist *, cpl_bivector *, 
        int *) ;
static cpl_bivector * hawki_cal_illum_find_pos(cpl_imagelist *, 
        cpl_bivector *, int) ;

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static struct {
    /* Inputs */
    int         subtract ;
    int         degree ;
    int         s_hx ;
    int         s_hy ;
    double      star_r ;
    double      bg_r1 ;
    double      bg_r2 ;
} hawki_cal_illum_config ;

static char hawki_cal_illum_description[] =
"hawki_cal_illum -- Illumination recipe\n"
"The files listed in the Set Of Frames (sof-file) must be tagged:\n"
"raw-file.fits "HAWKI_CAL_ILLUM_RAW" or\n"
"flat-file.fits "HAWKI_CALPRO_FLAT" or\n"
"bpm-file.fits "HAWKI_CALPRO_BPM"\n" ;

/*-----------------------------------------------------------------------------
                                Functions code
 -----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok, -1 otherwise

  Create the recipe instance and make it available to the application using the 
  interface. This function is exported.
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist * list)
{
    cpl_recipe  *   recipe = cpl_calloc(1, sizeof *recipe ) ;
    cpl_plugin  *   plugin = &recipe->interface ;

    cpl_plugin_init(plugin,
                    CPL_PLUGIN_API,
                    HAWKI_BINARY_VERSION,
                    CPL_PLUGIN_TYPE_RECIPE,
                    "hawki_cal_illum",
                    "Illumination recipe",
                    hawki_cal_illum_description,
                    "Yves Jung",
                    PACKAGE_BUGREPORT,  
                    hawki_get_license(),
                    hawki_cal_illum_create,
                    hawki_cal_illum_exec,
                    hawki_cal_illum_destroy) ;

    cpl_pluginlist_append(list, plugin) ;
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int hawki_cal_illum_create(cpl_plugin * plugin)
{
    cpl_recipe      *   recipe ;
    cpl_parameter   *   p ;
        
    /* Check that the plugin is part of a valid recipe */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE) 
        recipe = (cpl_recipe *)plugin ;
    else return -1 ;

    /* Create the parameters list in the cpl_recipe object */
    recipe->parameters = cpl_parameterlist_new() ; 

    /* Fill the parameters list */
    /* --subtract */
    p = cpl_parameter_new_value("hawki.hawki_cal_illum.subtract",
            CPL_TYPE_BOOL, "Flag to subtract the next images",
            "hawki.hawki_cal_illum", TRUE) ;
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "subtract") ;
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV) ;
    cpl_parameterlist_append(recipe->parameters, p) ;
    /* --degree */
    p = cpl_parameter_new_value("hawki.hawki_cal_illum.degree",
            CPL_TYPE_INT, "Polynomial degree for the fit",
            "hawki.hawki_cal_illum", 3) ;
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "degree") ;
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV) ;
    cpl_parameterlist_append(recipe->parameters, p) ;
       /* --star_r */
    p = cpl_parameter_new_value("hawki.hawki_cal_illum.star_r", CPL_TYPE_DOUBLE,
            "the star radius", "hawki.hawki_cal_illum", -1.0) ;
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "star_r") ;
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV) ;
    cpl_parameterlist_append(recipe->parameters, p) ;
    /* --bg_r1 */
    p = cpl_parameter_new_value("hawki.hawki_cal_illum.bg_r1", CPL_TYPE_DOUBLE,
            "the internal background radius", "hawki.hawki_cal_illum", -1.0) ;
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "bg_r1") ;
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV) ;
    cpl_parameterlist_append(recipe->parameters, p) ;
    /* --bg_r2 */
    p = cpl_parameter_new_value("hawki.hawki_cal_illum.bg_r2", CPL_TYPE_DOUBLE,
            "the external background radius", "hawki.hawki_cal_illum", -1.0) ;
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "bg_r2") ;
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV) ;
    cpl_parameterlist_append(recipe->parameters, p) ;
    /* --s_hx */
    p = cpl_parameter_new_value("hawki.hawki_cal_illum.s_hx", CPL_TYPE_INT,
            "Half size of the seach box in x", "hawki.hawki_cal_illum", 50) ;
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "s_hx") ;
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV) ;
    cpl_parameterlist_append(recipe->parameters, p) ;
    /* --s_hy */
    p = cpl_parameter_new_value("hawki.hawki_cal_illum.s_hy", CPL_TYPE_INT,
            "Half size of the seach box in y", "hawki.hawki_cal_illum", 50) ;
    cpl_parameter_set_alias(p, CPL_PARAMETER_MODE_CLI, "s_hy") ;
    cpl_parameter_disable(p, CPL_PARAMETER_MODE_ENV) ;
    cpl_parameterlist_append(recipe->parameters, p) ;
 
    /* Return */
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int hawki_cal_illum_exec(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe ;
    
    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE) 
        recipe = (cpl_recipe *)plugin ;
    else return -1 ;

    /* Issue a banner */
    hawki_print_banner();

    return hawki_cal_illum(recipe->parameters, recipe->frames) ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int hawki_cal_illum_destroy(cpl_plugin * plugin)
{
    cpl_recipe  *   recipe ;
    
    /* Get the recipe out of the plugin */
    if (cpl_plugin_get_type(plugin) == CPL_PLUGIN_TYPE_RECIPE) 
        recipe = (cpl_recipe *)plugin ;
    else return -1 ;

    cpl_parameterlist_delete(recipe->parameters) ; 
    return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    parlist     the parameters list
  @param    frameset   the frames list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int hawki_cal_illum(
        cpl_parameterlist   *   parlist, 
        cpl_frameset        *   frameset)
{
    cpl_parameter       *   par ;
    const char          *   flat ;
    const char          *   bpm ;
    cpl_frameset        *   rawframes ;
    cpl_frameset        *   cur_frames ;
    int                 *   labels ;
    cpl_image           *   illum[HAWKI_NB_DETECTORS] ;
    cpl_table           *   flux[HAWKI_NB_DETECTORS] ;
    int                     i ;

    /* Initialise */
    rawframes = NULL ;

    /* Retrieve input parameters */
    par = cpl_parameterlist_find(parlist, "hawki.hawki_cal_illum.subtract") ;
    hawki_cal_illum_config.subtract = cpl_parameter_get_bool(par) ;
    par = cpl_parameterlist_find(parlist, "hawki.hawki_cal_illum.degree") ;
    hawki_cal_illum_config.degree = cpl_parameter_get_int(par) ;
    par = cpl_parameterlist_find(parlist, "hawki.hawki_cal_illum.s_hx") ;
    hawki_cal_illum_config.s_hx = cpl_parameter_get_int(par) ;
    par = cpl_parameterlist_find(parlist, "hawki.hawki_cal_illum.s_hy") ;
    hawki_cal_illum_config.s_hy = cpl_parameter_get_int(par) ;
    par = cpl_parameterlist_find(parlist, "hawki.hawki_cal_illum.star_r") ;
    hawki_cal_illum_config.star_r = cpl_parameter_get_double(par) ;
    par = cpl_parameterlist_find(parlist, "hawki.hawki_cal_illum.bg_r1") ;
    hawki_cal_illum_config.bg_r1 = cpl_parameter_get_double(par) ;
    par = cpl_parameterlist_find(parlist, "hawki.hawki_cal_illum.bg_r2") ;
    hawki_cal_illum_config.bg_r2 = cpl_parameter_get_double(par) ;
 
    /* Identify the RAW and CALIB frames in the input frameset */
    if (hawki_dfs_set_groups(frameset)) {
        cpl_msg_error(__func__, "Cannot identify RAW and CALIB frames") ;
        return -1 ;
    }

    /* Retrieve calibration data */
    flat = hawki_extract_first_filename(frameset, HAWKI_CALPRO_FLAT) ;
    bpm = hawki_extract_first_filename(frameset, HAWKI_CALPRO_BPM) ;

    /* Retrieve raw frames */
    if ((rawframes = hawki_extract_frameset(frameset, 
                    HAWKI_CAL_ILLUM_RAW)) == NULL) {
        cpl_msg_error(__func__, "No raw frame in input") ;
        return -1 ;
    }

    /* Labelise frames */
    cpl_msg_info(__func__, "Labelise the frames") ;
    if ((labels = hawki_detectors_labelise(rawframes)) == NULL) {
        cpl_msg_error(__func__, "Cannot labelise") ;
        cpl_frameset_delete(rawframes) ;
        return -1 ;
    }

    /* Loop on the chips */
    for (i=0 ; i<HAWKI_NB_DETECTORS ; i++) {
        cpl_msg_info(__func__, "Reduce chip nb %d", i+1) ;
        
        /* Initialise */
        illum[i] = NULL ;

        /* Get the frames for which the star is in chip number i+1 */
        if ((cur_frames = cpl_frameset_extract(rawframes, labels,
                        i+1)) == NULL) {
            cpl_msg_warning(__func__, "No frames for chip nb %d", i+1) ;
            /* Create the flux table */
            flux[i] = cpl_table_new(1) ;
            cpl_table_new_column(flux[i],HAWKI_COL_ILLUM_POSX, CPL_TYPE_DOUBLE);
            cpl_table_new_column(flux[i],HAWKI_COL_ILLUM_POSY, CPL_TYPE_DOUBLE);
            cpl_table_new_column(flux[i],HAWKI_COL_ILLUM_FLUX, CPL_TYPE_DOUBLE);
        } else {
            /* Create the flux table */
            flux[i] = cpl_table_new(cpl_frameset_get_size(cur_frames)) ;
            cpl_table_new_column(flux[i],HAWKI_COL_ILLUM_POSX, CPL_TYPE_DOUBLE);
            cpl_table_new_column(flux[i],HAWKI_COL_ILLUM_POSY, CPL_TYPE_DOUBLE);
            cpl_table_new_column(flux[i],HAWKI_COL_ILLUM_FLUX, CPL_TYPE_DOUBLE);

            /* Get the illumination */
            if ((illum[i] = hawki_cal_illum_compute(cur_frames, i+1, 
                            bpm, flat, &(flux[i]))) == NULL) {
                cpl_msg_warning(__func__, "Cannot compute for chip %d", i+1) ;
                cpl_error_reset() ;
            }
            cpl_frameset_delete(cur_frames) ;
        }
    }
    cpl_frameset_delete(rawframes) ;
    cpl_free(labels) ;
    
    /* Save the product */
    cpl_msg_info(__func__, "Save the product") ;
    cpl_msg_indent_more() ;
    if (hawki_cal_illum_save((const cpl_image **)illum, 
                (const cpl_table **)flux, parlist, frameset)) {
        cpl_msg_error(__func__, "Cannot save the product") ;
        for (i=0 ; i<HAWKI_NB_DETECTORS ; i++) cpl_table_delete(flux[i]) ;
        for (i=0 ; i<HAWKI_NB_DETECTORS ; i++) cpl_image_delete(illum[i]) ;
        cpl_msg_indent_less() ;
        return -1 ;
    }
    for (i=0 ; i<HAWKI_NB_DETECTORS ; i++) cpl_table_delete(flux[i]) ;
    for (i=0 ; i<HAWKI_NB_DETECTORS ; i++) cpl_image_delete(illum[i]) ;
    cpl_msg_indent_less() ;

    /* Return */
    if (cpl_error_get_code()) return -1 ;
    else return 0 ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    The recipe data reduction part is implemented here 
  @param    raw         the raw frames
  @param    chip        the current chip
  @param    bpm         the bad pixels map or NULL
  @param    flat        the flat of NULL
  @param    fl          the returned flux table
  @return   the illumination image or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_image * hawki_cal_illum_compute(
        cpl_frameset    *   raw,
        int                 chip,
        const char      *   bpm,
        const char      *   flat,
        cpl_table       **  fl)
{
    cpl_bivector    *   offsets ;
    cpl_imagelist   *   in ;
    cpl_image       *   ima1 ;
    cpl_image       *   ima2 ;
    cpl_bivector    *   positions ;
    cpl_bivector    *   purged_pos ;
    double          *   positions_x,
                    *   positions_y,
                    *   purged_pos_x,
                    *   purged_pos_y ;
    cpl_vector      *   flux ;
    cpl_vector      *   purged_flux ;
    cpl_polynomial  *   poly ;
    cpl_image       *   poly_ima ;
    int                 nval, nx, ny ;
    double              val ;
    int                 ival ;
    int                 i, j ;
    
    /* Check entries */
    if (chip < 1 || chip > HAWKI_NB_DETECTORS) return NULL ;
    if (raw == NULL) return NULL ;
    if (fl == NULL) return NULL ;
         
    /* Get the offsets from the raw frames */
    cpl_msg_info(__func__, "Get the offsets from the files headers") ;
    cpl_msg_indent_more() ;
    if ((offsets = hawki_get_header_tel_offsets(raw)) == NULL) {
        cpl_msg_error(__func__, "Cannot get the offsets") ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    cpl_msg_indent_less() ;

    /* Load the input data */
    cpl_msg_info(__func__, "Load the input data") ;
    cpl_msg_indent_more() ;
    if ((in = hawki_load_detector(raw, chip, CPL_TYPE_FLOAT)) == NULL) {
        cpl_msg_error(__func__, "Cannot load input data") ;
        cpl_bivector_delete(offsets) ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    cpl_msg_indent_less() ;

    /* Apply the calibrations */
    if (flat || bpm) {
        cpl_msg_indent_more() ;
        if (hawki_flat_bpm_detector_calib(in, flat, bpm) == -1) {
            cpl_msg_error(__func__, "Cannot apply calibrations") ;
            cpl_bivector_delete(offsets) ;
            cpl_imagelist_delete(in) ;
            cpl_msg_indent_less() ;
            return NULL ;
        }
        cpl_msg_indent_less() ;
    }

    /* Subtract the consecutive frames */
    if (hawki_cal_illum_config.subtract) {
        cpl_msg_info(__func__, "Subtract the next images") ;
        ima1 = cpl_image_duplicate(cpl_imagelist_get(in, 0)) ;
        for (i=0 ; i<cpl_imagelist_get_size(in) ; i++) {
            if (i==cpl_imagelist_get_size(in)-1) {
                ima2 = ima1 ;
            } else {
                ima2 = cpl_imagelist_get(in, i+1) ;
            }
            cpl_image_subtract(cpl_imagelist_get(in, i), ima2) ;
        }
        cpl_image_delete(ima1) ;
    }

    /* Find the positions */
    cpl_msg_info(__func__, "Find the positions") ;
    cpl_msg_indent_more() ;
    if ((positions=hawki_cal_illum_find_pos(in, offsets, chip)) == NULL) {
        cpl_msg_error(__func__, "Cannot find the positions") ;
        cpl_imagelist_delete(in) ;
        cpl_bivector_delete(offsets) ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    cpl_bivector_delete(offsets) ;
    cpl_msg_indent_less() ;

    /* Compute the photometry */
    cpl_msg_info(__func__, "Compute the photometry") ;
    cpl_msg_indent_more() ;
    if ((flux=hawki_cal_illum_phot(in, positions, &nval)) == NULL) {
        cpl_msg_error(__func__, "Cannot find the positions") ;
        cpl_imagelist_delete(in) ;
        cpl_bivector_delete(positions) ;
        cpl_msg_indent_less() ;
        return NULL ;
    }
    nx = cpl_image_get_size_x(cpl_imagelist_get(in, 0)) ;
    ny = cpl_image_get_size_y(cpl_imagelist_get(in, 0)) ;
    cpl_imagelist_delete(in) ;
    cpl_msg_indent_less() ;
    if (nval < 1) {
        cpl_msg_error(__func__, "No flux computed") ;
        cpl_bivector_delete(positions) ;
        cpl_vector_delete(flux) ;
        return NULL ;
    }
    
    /* Purge positions */
    purged_pos = cpl_bivector_new(nval) ;
    purged_pos_x = cpl_bivector_get_x_data(purged_pos) ;
    purged_pos_y = cpl_bivector_get_y_data(purged_pos) ;
    positions_x = cpl_bivector_get_x_data(positions) ;
    positions_y = cpl_bivector_get_y_data(positions) ;
    purged_flux = cpl_vector_new(nval) ;
    j = 0 ;
    for (i=0 ; i<cpl_vector_get_size(flux) ; i++) {
        if (fabs(cpl_vector_get(flux, i)) > 0) {
            purged_pos_x[j] = positions_x[i] ;
            purged_pos_y[j] = positions_y[i] ;
            cpl_vector_set(purged_flux, j, cpl_vector_get(flux, i)) ;
            j++ ;
        }
    }

    /* Fill the flux table */
    for (i=0 ; i<cpl_vector_get_size(flux) ; i++) {
        cpl_table_set_double(*fl, HAWKI_COL_ILLUM_POSX, i, positions_x[i]) ;
        cpl_table_set_double(*fl, HAWKI_COL_ILLUM_POSY, i, positions_y[i]) ;
        cpl_table_set_double(*fl, HAWKI_COL_ILLUM_FLUX, i, 
                cpl_vector_get(flux, i)) ;
    }
    cpl_bivector_delete(positions) ;
    cpl_vector_delete(flux) ;
    
    /* Fit the polynomial */
    cpl_msg_info(__func__, "Compute the polynomial") ;
    if ((poly = cpl_polynomial_fit_2d_create(purged_pos, purged_flux,
            hawki_cal_illum_config.degree, NULL)) == NULL) {
        cpl_msg_error(__func__, "Cannot fit the polynomial") ;
        cpl_bivector_delete(purged_pos) ;
        cpl_vector_delete(purged_flux) ;
        return NULL ;
    }
    cpl_bivector_delete(purged_pos) ;
    cpl_vector_delete(purged_flux) ;
   
    /* Create the polynomial image */
    poly_ima = cpl_image_new(nx, ny, CPL_TYPE_FLOAT) ;
    cpl_image_fill_polynomial(poly_ima, poly, 1.0, 1.0, 1.0, 1.0) ;
    cpl_polynomial_delete(poly) ;

    /* Normalise the image */
    val = cpl_image_get(poly_ima, 1024, 1024, &ival) ;
    cpl_image_divide_scalar(poly_ima, val) ;

    return poly_ima ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Compute the photometry of the bright star in the images.
  @param    ilist       the images list
  @param    positions   the star positions
  @param    nvalid      number of valid points
  @return   the flux in a vector or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_vector * hawki_cal_illum_phot(
        cpl_imagelist   *   ilist,
        cpl_bivector    *   positions,
        int             *   nvalid)
{
    cpl_vector          *   flux ;
    cpl_image           *   cur_ima ;
    double              *   pos_x ;
    double              *   pos_y ;
    int                     ni ;
    int                     nok ;
    double                  bg, fl ;
    cpl_bivector        *   iqe_res ;
    double                  fwhm_x, fwhm_y, r, r1, r2 ;
    int                     i ;

    /* Test entries */
    if (ilist == NULL || positions == NULL || nvalid == NULL) return NULL ;
    ni = cpl_imagelist_get_size(ilist) ;
    pos_x = cpl_bivector_get_x_data(positions) ;
    pos_y = cpl_bivector_get_y_data(positions) ;
    nok = 0 ;

    /* Create the flux vector */
    flux = cpl_vector_new(ni) ;

    /* Loop on the frames */
    for (i=0 ; i<ni ; i++) {
        cur_ima = cpl_imagelist_get(ilist, i) ;

        /* FWHM_X / FWHM_Y */
        if ((iqe_res = cpl_image_iqe(cur_ima,
                        (int)(pos_x[i]-10.0), (int)(pos_y[i]-10.0),
                        (int)(pos_x[i]+10.0), (int)(pos_y[i]+10.0))) == NULL) {
            cpl_msg_debug(__func__,"Cannot get FWHM for image %d", i+1);
            fwhm_x = fwhm_y = -1.0 ;
            cpl_error_reset() ;
        } else {
            fwhm_x = cpl_vector_get(cpl_bivector_get_x(iqe_res), 2) ;
            fwhm_y = cpl_vector_get(cpl_bivector_get_x(iqe_res), 3) ;
            cpl_bivector_delete(iqe_res) ;
        }

        /* Determine the radii */
        r = hawki_cal_illum_config.star_r ;
        if (r < 0) {
            if (fwhm_x>0 && fwhm_y>0)   r = 5*(fwhm_x+fwhm_y)/2.0 ;
            else                        r = HAWKI_PHOT_STAR_RADIUS ;
        }
        r1 = hawki_cal_illum_config.bg_r1 ;
        r2 = hawki_cal_illum_config.bg_r2 ;
        if (r1 < 0) r1 = r + 10.0 ;
        if (r2 < 0) r2 = r1 + 20.0 ;

        bg = irplib_strehl_ring_background(cur_ima, (int)pos_x[i],
                (int)pos_y[i], r1, r2, IRPLIB_BG_METHOD_MEDIAN) ;
        fl = irplib_strehl_disk_flux(cur_ima, (int)pos_x[i], (int)pos_y[i], 
                r, bg) ;
        cpl_vector_set(flux, i, fl); 
        if (fabs(fl)>1e-3) nok++ ; 
        cpl_msg_info(__func__, "Flux for image %d: %g (r=%g, r1=%g, r2=%g)", 
                i+1, fl, r, r1, r2) ;
    }
    *nvalid = nok ;
    return flux ;
}
 
/*----------------------------------------------------------------------------*/
/**
  @brief    Get the positions of the bright star in the images.
  @param    ilist       the images list
  @param    offsets     the offsets between frames
  @param    chip        the current chip
  @return   the positions or NULL in error case
 */
/*----------------------------------------------------------------------------*/
static cpl_bivector * hawki_cal_illum_find_pos(
        cpl_imagelist   *   ilist,
        cpl_bivector    *   offsets,
        int                 chip)
{
    cpl_bivector        *   positions ;
    double              *   positions_x ;
    double              *   positions_y ;
    double              *   offsets_x ;
    double              *   offsets_y ;
    cpl_image           *   tmp_ima ;
    cpl_image           *   cur_ima ;
    cpl_image           *   filtered_ima ;
    cpl_matrix          *   kernel ;
    int                     nx, ny, ni ;
    int                     llx, lly, urx, ury, posx, posy, gap ;
    int                     i ;

    /* Test entries */
    if (ilist == NULL || offsets == NULL) return NULL ;
    if (chip < 1 || chip > HAWKI_NB_DETECTORS) return NULL ;
    ni = cpl_imagelist_get_size(ilist) ;
    gap = 147 ;

    /* Create the positions bivector */
    positions = cpl_bivector_new(ni) ;
    positions_x = cpl_bivector_get_x_data(positions) ;
    positions_y = cpl_bivector_get_y_data(positions) ;
    offsets_x = cpl_bivector_get_x_data(offsets) ;
    offsets_y = cpl_bivector_get_y_data(offsets) ;

    /* Create the kernel */
    kernel = cpl_matrix_new(3, 3) ;
    cpl_matrix_fill(kernel, 1.0) ;
    
    /* Loop on all the images */
    for (i=0 ; i<ni ; i++) {
        cur_ima = cpl_imagelist_get(ilist, i) ;
        nx = cpl_image_get_size_x(cur_ima) ;
        ny = cpl_image_get_size_y(cur_ima) ;

        /* Define zone */
        if (chip == 1) {
            llx = nx + gap/2.0 + offsets_x[i] - hawki_cal_illum_config.s_hx ;
            urx = nx + gap/2.0 + offsets_x[i] + hawki_cal_illum_config.s_hx ;
            lly = ny + gap/2.0 + offsets_y[i] - hawki_cal_illum_config.s_hy ;
            ury = ny + gap/2.0 + offsets_y[i] + hawki_cal_illum_config.s_hy ;
        } else if (chip == 2) {
            llx = offsets_x[i] - gap/2.0 - hawki_cal_illum_config.s_hx ;
            urx = offsets_x[i] - gap/2.0 + hawki_cal_illum_config.s_hx ;
            lly = ny + gap/2.0 + offsets_y[i] - hawki_cal_illum_config.s_hy ;
            ury = ny + gap/2.0 + offsets_y[i] + hawki_cal_illum_config.s_hy ;
        } else if (chip == 3) {
            llx = offsets_x[i] - gap/2.0 - hawki_cal_illum_config.s_hx ;
            urx = offsets_x[i] - gap/2.0 + hawki_cal_illum_config.s_hx ;
            lly = offsets_y[i] - gap/2.0 - hawki_cal_illum_config.s_hy ;
            ury = offsets_y[i] - gap/2.0 + hawki_cal_illum_config.s_hy ;
        } else if (chip == 4) {
            llx = nx + gap/2.0 + offsets_x[i] - hawki_cal_illum_config.s_hx ;
            urx = nx + gap/2.0 + offsets_x[i] + hawki_cal_illum_config.s_hx ;
            lly = offsets_y[i] - gap/2.0 - hawki_cal_illum_config.s_hy ;
            ury = offsets_y[i] - gap/2.0 + hawki_cal_illum_config.s_hy ;
        }

        /* Test zone */
        if (llx>urx || lly>ury || llx<1 || urx>nx || lly<1 || ury>ny) {
            cpl_msg_error(__func__, "Bad specified zone") ;
            cpl_bivector_delete(positions) ;
            cpl_matrix_delete(kernel) ;
            return NULL ;
        }
        
        /* Extract */
        tmp_ima = cpl_image_extract(cur_ima, llx, lly, urx, ury) ;
        filtered_ima =cpl_image_filter_median(tmp_ima, kernel) ;
        cpl_image_delete(tmp_ima) ;

        /* Find the max */
        cpl_image_get_maxpos(filtered_ima, &posx, &posy) ;
        positions_x[i] = llx + posx ;
        positions_y[i] = lly + posy ;
        cpl_image_delete(filtered_ima) ;
        cpl_msg_info(__func__, "Star found at pos %g, %g in image %d",
                positions_x[i], positions_y[i], i+1) ;
    }
    cpl_matrix_delete(kernel) ;

    return positions ;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Save the products
  @param    illum       the HAWKI_NB_DETECTORS illum images
  @param    flux        the HAWKI_NB_DETECTORS flux tables
  @param    parlist     the parameters list
  @param    set         the frames set
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int hawki_cal_illum_save(
        const cpl_image     **  illum,
        const cpl_table     **  flux,
        cpl_parameterlist   *   parlist,
        cpl_frameset        *   set)
{
    cpl_propertylist    **  qclists ;
    cpl_propertylist    *   paflist ;
    cpl_propertylist    *   paflist_cur ;
    const cpl_frame     *   ref_frame ;
    cpl_propertylist    *   inputlist ;
    int                     ext_nb ;
    char                *   filename ;
    char                    sval[16] ;
    const char          *   recipe_name = "hawki_cal_illum" ;
    int                     i;

    /* Get the reference frame */
    ref_frame = irplib_frameset_get_first_from_group(set, CPL_FRAME_GROUP_RAW) ;

    /* Create the QC lists */
    qclists = cpl_malloc(HAWKI_NB_DETECTORS * sizeof(cpl_propertylist*)) ;
    for (i=0 ; i<HAWKI_NB_DETECTORS ; i++) {
        qclists[i] = cpl_propertylist_new() ;
        cpl_propertylist_append_double(qclists[i], "ESO QC DATANCOM",
                cpl_frameset_get_size(set)) ;

        /* Propagate some keywords from input raw frame extensions */
        ext_nb=hawki_get_ext_from_detector(cpl_frame_get_filename(ref_frame), i+1);
        inputlist = cpl_propertylist_load_regexp(
                cpl_frame_get_filename(ref_frame), ext_nb,
                HAWKI_HEADER_EXT_FORWARD, 0) ;
        cpl_propertylist_copy_property_regexp(qclists[i], inputlist, "", 0) ;
        cpl_propertylist_delete(inputlist) ;
    }

    /* Write the image */
    hawki_images_save(set,
                      parlist,
                      set,
                      (const cpl_image **)illum, 
                      recipe_name,
                      HAWKI_CALPRO_ILLUM, 
                      HAWKI_PROTYPE_ILLUM,
                      NULL,
                      (const cpl_propertylist**)qclists,
                      "hawki_cal_illum.fits") ;
    /* To handle cases where a chip is missing */
    if (cpl_error_get_code() == CPL_ERROR_NULL_INPUT) cpl_error_reset() ;

    /* Write the flux table */
    hawki_tables_save(set,
                      parlist,
                      set,
                      (const cpl_table **)flux,
                      recipe_name,
                      HAWKI_CALPRO_ILLUM_PHOTOM,
                      HAWKI_PROTYPE_ILLUM_PHOTOM,
                      NULL,
                      (const cpl_propertylist**)qclists,
                      "hawki_cal_illum_photom.fits") ;
    /* To handle cases where a chip is missing */
    if (cpl_error_get_code() == CPL_ERROR_NULL_INPUT) cpl_error_reset() ;
   
    /* Remove the keywords for the FITS extensions */
    for (i=0 ; i<HAWKI_NB_DETECTORS ; i++) {
        cpl_propertylist_erase_regexp(qclists[i], HAWKI_HEADER_EXT_FORWARD, 0) ;
    }

    /* Get FITS header from reference file */
    paflist = cpl_propertylist_load_regexp(cpl_frame_get_filename(ref_frame),
            0, HAWKI_HEADER_PRI_TOPAF, 0) ;

    /* Add the PRO REC1 ID in the paf files */
    cpl_propertylist_prepend_string(paflist, "PRO REC1 ID", recipe_name) ;

    /* Add the PRO CATG keyword. This is a workaround, since the current 
     * implemented behaviour is not to have PAF files associated to products.
     * All this will be removed with the FITS2PAF utility in esorex 
     */
    cpl_propertylist_prepend_string(paflist, "PRO CATG", HAWKI_CALPRO_ILLUM);

    /* Write the extension PAF files */
    for (i=0 ; i<HAWKI_NB_DETECTORS ; i++) {
        /* Duplicate paflist */
        paflist_cur = cpl_propertylist_duplicate(paflist) ;

        /* Add the EXTNAME keyword in the PAF */
        sprintf(sval, "CHIP%d.INT1", i+1) ;
        cpl_propertylist_prepend_string(paflist_cur, "EXTNAME", sval) ;

        /* Get FITS header from reference file extension */
        ext_nb=hawki_get_ext_from_detector(cpl_frame_get_filename(ref_frame), i+1);
        inputlist = cpl_propertylist_load_regexp(
                cpl_frame_get_filename(ref_frame), ext_nb,
                HAWKI_HEADER_EXT_TOPAF, 0) ;
        cpl_propertylist_copy_property_regexp(paflist_cur, inputlist, "", 0) ;
        cpl_propertylist_delete(inputlist) ;

        /* End of Workaround */

        /* Copy the QC in paflist */
        cpl_propertylist_copy_property_regexp(paflist_cur, qclists[i], "", 0) ;

        /* Paf file name */
        filename = cpl_sprintf("%s_%02d.paf", recipe_name, i+1) ;

        /* Save the PAF file */
        cpl_dfs_save_paf("HAWKI",
                recipe_name,
                paflist_cur,
                filename) ;
        cpl_free(filename) ;
        cpl_propertylist_delete(paflist_cur) ;
    }

    /* Free and return */
    for (i=0 ; i<HAWKI_NB_DETECTORS ; i++) {
        cpl_propertylist_delete(qclists[i]) ;
    }
    cpl_free(qclists) ;
    cpl_propertylist_delete(paflist) ;
    return  0;
}

